# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


TRACKING_PREFIX = {
    "AL": "Agentes de leitura",
    "AR": "Avisos de recebimento",
    "AS": "PAC - Ação Social",
    "CA": "Encomenda Internacional - Colis",
    "CB": "Encomenda Internacional - Colis",
    "CC": "Encomenda Internacional - Colis",
    "CD": "Encomenda Internacional - Colis",
    "CE": "Encomenda Internacional - Colis",
    "CF": "Encomenda Internacional - Colis",
    "CG": "Encomenda Internacional - Colis",
    "CH": "Encomenda Internacional - Colis",
    "CI": "Encomenda Internacional - Colis",
    "CJ": "Encomenda Internacional - Colis",
    "CK": "Encomenda Internacional - Colis",
    "CL": "Encomenda Internacional - Colis",
    "CM": "Encomenda Internacional - Colis",
    "CN": "Encomenda Internacional - Colis",
    "CO": "Encomenda Internacional - Colis",
    "CP": "Encomenda Internacional - Colis",
    "CQ": "Encomenda Internacional - Colis",
    "CR": "Carta registrada sem Valor Declarado",
    "CS": "Encomenda Internacional - Colis",
    "CT": "Encomenda Internacional - Colis",
    "CU": "Encomenda internacional - Colis",
    "CV": "Encomenda Internacional - Colis",
    "CW": "Encomenda Internacional - Colis",
    "CX": "Encomenda internacional - Colis ou Selo Lacre para Caixetas",
    "CY": "Encomenda Internacional - Colis",
    "CZ": "Encomenda Internacional - Colis",
    "DA": "SEDEX ou Remessa Expressa com AR Digital",
    "DB": "SEDEX ou Remessa Expressa com AR Digital (Bradesco)",
    "DC": "Remessa Expressa CRLV/CRV/CNH e Notificações",
    "DD": "Devolução de documentos",
    "DE": "Remessa Expressa Talão/Cartão com AR",
    "DF": "e-SEDEX",
    "DG": "SEDEX",
    "DI": "SEDEX ou Remessa Expressa com AR Digital (Itaú)",
    "DJ": "SEDEX",
    "DK": "PAC Extra Grande",
    "DL": "SEDEX",
    "DM": "e-SEDEX",
    "DN": "SEDEX",
    "DO": "SEDEX ou Remessa Expressa com AR Digital (Itaú)",
    "DP": "SEDEX Pagamento na Entrega",
    "DQ": "SEDEX ou Remessa Expressa com AR Digital (Santander)",
    "DR": "Remessa Expressa com AR Digital (Santander)",
    "DS": "SEDEX ou Remessa Expressa com AR Digital (Santander)",
    "DT": "Remessa econômica com AR Digital (DETRAN)",
    "DU": "e-SEDEX",
    "DV": "SEDEX c/ AR digital",
    "DW": "Encomenda SEDEX (Etiqueta Lógica)",
    "DX": "SEDEX 10",
    "EA": "Encomenda Internacional - EMS",
    "EB": "Encomenda Internacional - EMS",
    "EC": "PAC",
    "ED": "Packet Express",
    "EE": "Encomenda Internacional - EMS",
    "EF": "Encomenda Internacional - EMS",
    "EG": "Encomenda Internacional - EMS",
    "EH": "Encomenda Internacional - EMS ou Encomenda com AR Digital",
    "EI": "Encomenda Internacional - EMS",
    "EJ": "Encomenda Internacional - EMS",
    "EK": "Encomenda Internacional - EMS",
    "EL": "Encomenda Internacional - EMS",
    "EM": "Encomenda Internacional - SEDEX Mundi",  # BR Suffix
    # "EM": "Encomenda Internacional - EMS Importação",
    "EN": "Encomenda Internacional - EMS",
    "EO": "Encomenda Internacional - EMS",
    "EP": "Encomenda Internacional - EMS",
    "EQ": "Encomenda de serviço não expressa (ECT)",
    "ER": "Objeto registrado",
    "ES": "e-SEDEX ou EMS",
    "ET": "Encomenda Internacional - EMS",
    "EU": "Encomenda Internacional - EMS",
    "EV": "Encomenda Internacional - EMS",
    "EW": "Encomenda Internacional - EMS",
    "EX": "Encomenda Internacional - EMS",
    "EY": "Encomenda Internacional - EMS",
    "EZ": "Encomenda Internacional - EMS",
    "FA": "FAC registrado",
    "FE": "Encomenda FNDE",
    "FF": "Objeto registrado (DETRAN)",
    "FH": "FAC registrado com AR Digital",
    "FM": "FAC monitorado",
    "FR": "FAC registrado",
    "IA": "Logística Integrada (agendado / avulso)",
    "IC": "Logística Integrada (a cobrar)",
    "ID": "Logística Integrada (devolução de documento)",
    "IE": "Logística Integrada (Especial)",
    "IF": "CPF",
    "II": "Logística Integrada (ECT)",
    "IK": "Logística Integrada com Coleta Simultânea",
    "IM": "Logística Integrada (Medicamentos)",
    "IN": "Correspondência e EMS recebido do Exterior",
    "IP": "Logística Integrada (Programada)",
    "IR": "Impresso Registrado",
    "IS": "Logística integrada standard (medicamentos)",
    "IT": "Remessa Expressa Medicamentos / Logística Integrada Termolábil",
    "IU": "Logística Integrada (urgente)",
    "IX": "EDEI Expresso",
    "JA": "Remessa econômica com AR Digital",
    "JB": "Remessa econômica com AR Digital",
    "JC": "Remessa econômica com AR Digital",
    "JD": "Remessa econômica Talão/Cartão",
    "JE": "Remessa econômica com AR Digital",
    "JF": "Remessa econômica com AR Digital",
    "JG": "Objeto registrado urgente/prioritário",
    "JH": "Objeto registrado urgente / prioritário",
    "JI": "Remessa econômica Talão/Cartão",
    "JJ": "Objeto registrado (Justiça)",
    "JK": "Remessa econômica Talão/Cartão",
    "JL": "Objeto registrado",
    "JM": "Mala Direta Postal Especial",
    "JN": "Objeto registrado econômico",
    "JO": "Objeto registrado urgente",
    "JP": "Receita Federal",
    "JQ": "Remessa econômica com AR Digital",
    "JR": "Objeto registrado urgente / prioritário",
    "JS": "Objeto registrado",
    "JT": "Objeto Registrado Urgente",
    "JV": "Remessa Econômica (c/ AR DIGITAL)",
    "LA": "SEDEX com Logística Reversa Simultânea em Agência",
    "LB": "e-SEDEX com Logística Reversa Simultânea em Agência",
    "LC": "Objeto Internacional (Prime)",
    "LE": "Logística Reversa Econômica",
    "LF": "Objeto Internacional (Prime)",
    "LI": "Objeto Internacional (Prime)",
    "LJ": "Objeto Internacional (Prime)",
    "LK": "Objeto Internacional (Prime)",
    "LM": "Objeto Internacional (Prime)",
    "LN": "Objeto Internacional (Prime)",
    "LP": "PAC com Logística Reversa Simultânea em Agência",
    "LS": "SEDEX Logística Reversa",
    "LV": "Logística Reversa Expressa",
    "LX": "Packet Standard / Econômica",
    "LZ": "Objeto Internacional (Prime)",
    "MA": "Serviços adicionais do Telegrama",
    "MB": "Telegrama (balcão)",
    "MC": "Telegrama (Fonado)",
    "MD": "SEDEX Mundi (Documento interno)",
    "ME": "Telegrama",
    "MF": "Telegrama (Fonado)",
    "MK": "Telegrama (corporativo)",
    "ML": "Fecha Malas (Rabicho)",
    "MM": "Telegrama (Grandes clientes)",
    "MP": "Telegrama (Pré-pago)",
    "MR": "AR digital",
    "MS": "Encomenda Saúde",
    "MT": "Telegrama (Telemail)",
    "MY": "Telegrama internacional (entrante)",
    "MZ": "Telegrama (Correios Online)",
    "NE": "Tele Sena resgatada",
    "NX": "EDEI Econômico (não urgente)",
    "PA": "Passaporte",
    "PB": "PAC",
    "PC": "PAC a Cobrar",
    "PD": "PAC",
    "PE": "PAC",
    "PF": "Passaporte",
    "PG": "PAC",
    "PH": "PAC",
    "PI": "PAC",
    "PJ": "PAC",
    "PK": "PAC Extra Grande",
    "PL": "PAC",
    "PR": "Reembolso Postal",
    "QQ": "Objeto de teste (SIGEP Web)",
    "RA": "Objeto registrado / prioritário",
    "RB": "Carta registrada",
    "RC": "Carta registrada com Valor Declarado",
    "RD": "Remessa econômica ou objeto registrado (DETRAN)",
    "RE": "Objeto registrado econômico",
    "RF": "Receita Federal",
    "RG": "Objeto registrado",
    "RH": "Objeto registrado com AR Digital",
    "RI": "Objeto registrado internacional prioritário",
    "RJ": "Objeto registrado",
    "RK": "Objeto registrado",
    "RL": "Objeto registrado",
    "RM": "Objeto registrado urgente",
    "RN": "Objeto registrado (SIGEPWEB ou Agência)",
    "RO": "Objeto registrado",
    "RP": "Reembolso Postal",
    "RQ": "Objeto registrado",
    "RR": "Objeto registrado",
    "RS": "Objeto registrado",
    "RT": "Remessa econômica Talão/Cartão",
    "RU": "Objeto registrado (ECT)",
    "RV": "Remessa econômica CRLV/CRV/CNH e Notificações com AR Digital",
    "RW": "Objeto internacional",
    "RX": "Objeto internacional",
    "RY": "Remessa econômica Talão/Cartão com AR Digital",
    "RZ": "Objeto registrado",
    "SA": "SEDEX",
    "SB": "SEDEX 10",
    "SC": "SEDEX a cobrar",
    "SD": "SEDEX ou Remessa Expressa (DETRAN)",
    "SE": "SEDEX",
    "SF": "SEDEX",
    "SG": "SEDEX",
    "SH": "SEDEX com AR Digital / SEDEX ou AR Digital",
    "SI": "SEDEX",
    "SJ": "SEDEX Hoje",
    "SK": "SEDEX",
    "SL": "SEDEX",
    "SM": "SEDEX 12",
    "SN": "SEDEX",
    "SO": "SEDEX",
    "SP": "SEDEX Pré-franqueado",
    "SQ": "SEDEX",
    "SR": "SEDEX",
    "SS": "SEDEX",
    "ST": "Remessa Expressa Talão/Cartão",
    "SU": "Encomenda de serviço expressa (ECT)",
    "SV": "Remessa Expressa CRLV/CRV/CNH e Notificações com AR Digital",
    "SW": "e-SEDEX",
    "SX": "SEDEX 10",
    "SY": "Remessa Expressa Talão/Cartão com AR Digital",
    "SZ": "SEDEX",
    "TC": "Objeto para treinamento",
    "TE": "Objeto para treinamento",
    "TS": "Objeto para treinamento",
    "VA": "Encomendas com valor declarado",
    "VC": "Encomendas",
    "VD": "Encomendas com valor declarado",
    "VE": "Encomendas",
    "VF": "Encomendas com valor declarado",
    "VV": "Objeto internacional",
    "XA": "Aviso de chegada (internacional)",
    "XM": "SEDEX Mundi",
    "XR": "Encomenda SUR Postal Expresso",
    "XX": "Encomenda SUR Postal 24 horas",
}

EXTRA_SERVICES = {
    1: {'code': "AR", 'name': "Aviso de Recebimento"},
    2: {'code': "MP", 'name': "Mão Própria Nacional"},
    19: {'code': "VD", 'name': "Valor Declarado (Encomendas)"},
    25: {'code': "RR", 'name': "Registro Nacional"},
}

EXTRA_SERVICE_AR = 1
EXTRA_SERVICE_MP = 2
EXTRA_SERVICE_VD = 19
EXTRA_SERVICE_RR = 25

SERVICES = {
    40215: {'id': 104707, 'description': 'SEDEX 10', 'category': 'SERVICO_COM_RESTRICAO', 'max_weight': 10000,
            'display_name': 'SEDEX 10', 'symbol': "premium", 'default_extra_services': [EXTRA_SERVICE_RR]},
    81019: {'id': 104672, 'description': 'E-SEDEX STANDARD', 'category': 'SERVICO_COM_RESTRICAO', 'max_weight': 15000,
            'display_name': 'E-SEDEX', 'symbol': "express", 'default_extra_services': [EXTRA_SERVICE_RR]},
    41068: {'id': 109819, 'description': 'PAC', 'category': 'PAC', 'display_name': 'PAC', 'max_weight': 30000,
            'symbol': "standard", 'default_extra_services': [EXTRA_SERVICE_RR]},
    40444: {'id': 109811, 'description': 'SEDEX - CONTRATO', 'category': 'SEDEX', 'max_weight': 30000,
            'display_name': 'SEDEX', 'symbol': "express", 'default_extra_services': [EXTRA_SERVICE_RR]},
    40436: {'id': 109810, 'description': 'SEDEX - CONTRATO', 'category': 'SEDEX', 'max_weight': 30000,
            'display_name': 'SEDEX', 'symbol': "express", 'default_extra_services': [EXTRA_SERVICE_RR]},
    40096: {'id': 104625, 'description': 'SEDEX (CONTRATO)', 'category': 'SEDEX', 'max_weight': 30000,
            'display_name': 'SEDEX', 'symbol': "express", 'default_extra_services': [EXTRA_SERVICE_RR]},
    40380: {'id': 109806, 'description': 'SEDEX REVERSO 40096', 'category': 'REVERSO', 'max_weight': 30000,
            'display_name': 'SEDEX', 'symbol': "express", 'default_extra_services': [EXTRA_SERVICE_RR]},
    40010: {'id': 104295, 'description': 'SEDEX A VISTA', 'category': 'SEDEX', 'max_weight': 30000,
            'display_name': 'SEDEX', 'symbol': "express", 'default_extra_services': [EXTRA_SERVICE_RR]},
    41211: {'id': 113546, 'description': 'PAC - CONTRATO', 'category': 'PAC', 'display_name': 'PAC',
            'max_weight': 30000, 'symbol': "standard", 'default_extra_services': [EXTRA_SERVICE_RR]},
    40630: {'id': 114976, 'description': 'SEDEX PAGAMENTO NA ENTREGA -', 'category': 'SEDEX', 'max_weight': 30000,
            'display_name': 'SEDEX', 'symbol': "express", 'default_extra_services': [EXTRA_SERVICE_RR]},
    40916: {'id': 118568, 'description': 'SEDEX AGRUPADO II', 'category': 'SEDEX', 'max_weight': 30000,
            'display_name': 'SEDEX', 'symbol': "express", 'default_extra_services': [EXTRA_SERVICE_RR]},
    40908: {'id': 118567, 'description': 'SEDEX AGRUPADO I', 'category': 'SEDEX', 'max_weight': 30000,
            'display_name': 'SEDEX', 'symbol': "express", 'default_extra_services': [EXTRA_SERVICE_RR]},
    41300: {'id': 120366, 'description': 'PAC GRANDES FORMATOS', 'category': 'SERVICO_COM_RESTRICAO',
            'max_weight': 50000, 'display_name': 'PAC', 'symbol': "standard",
            'default_extra_services': [EXTRA_SERVICE_RR]},
    40169: {'id': 115218, 'description': 'SEDEX 12', 'category': 'SERVICO_COM_RESTRICAO', 'max_weight': 10000,
            'display_name': 'SEDEX 12', 'symbol': "premium", 'default_extra_services': [EXTRA_SERVICE_RR]},
    40290: {'id': 108934, 'description': 'SEDEX HOJE', 'category': 'SERVICO_COM_RESTRICAO', 'max_weight': 10000,
            'display_name': 'SEDEX Hoje', 'symbol': "premium", 'default_extra_services': [EXTRA_SERVICE_RR]},
    10154: {'id': 118424, 'description': 'CARTA COMERCIAL  REGISTRADA', 'category': 'CARTA_REGISTRADA',
            'display_name': 'Carta Registrada'},
    41246: {'id': 115487, 'description': 'REM. CAMPANHA PAPAI NOEL DOS', 'category': 'SEM_CATEGORIA',
            'display_name': 'Papai Noel dos Correios'},
    40150: {'id': 115136, 'description': 'SERVICO DE PROTOCOLO POSTAL -', 'category': 'SEDEX',
            'display_name': 'Protocolo'},
    10065: {'id': 109480, 'description': 'CARTA COMERCIAL A FATURAR', 'category': 'CARTA_REGISTRADA',
            'display_name': 'Carta Comercial'}
}

SERVICE_PAC = 41068
SERVICE_SEDEX = 40096
SERVICE_SEDEX10 = 40215
SERVICE_SEDEX12 = 40169
SERVICE_E_SEDEX = 81019

REGIONAL_DIRECTIONS = {
    1: {'code': "AC", 'name': "AC - ADMINISTRAÇAO CENTRAL"},
    3: {'code': "ACR", 'name': "DR - ACRE"},
    4: {'code': "AL", 'name': "DR - ALAGOAS"},
    6: {'code': "AM", 'name': "DR - AMAZONAS"},
    5: {'code': "AP", 'name': "DR - AMAPÁ"},
    8: {'code': "BA", 'name': "DR - BAHIA"},
    10: {'code': "BSB", 'name': "DR - BRASÍLIA"},
    12: {'code': "CE", 'name': "DR - CEARÁ"},
    14: {'code': "ES", 'name': "DR - ESPIRITO SANTO"},
    16: {'code': "GO", 'name': "DR - GOIÁS"},
    18: {'code': "MA", 'name': "DR - MARANHÃO"},
    20: {'code': "MG", 'name': "DR - MINAS GERAIS"},
    22: {'code': "MS", 'name': "DR - MATO GROSSO DO SUL"},
    24: {'code': "MT", 'name': "DR - MATO GROSSO"},
    28: {'code': "PA", 'name': "DR - PARÁ"},
    30: {'code': "PB", 'name': "DR - PARAÍBA"},
    32: {'code': "PE", 'name': "DR - PERNAMBUCO"},
    34: {'code': "PI", 'name': "DR - PIAUÍ"},
    36: {'code': "PR", 'name': "DR - PARANÁ"},
    50: {'code': "RJ", 'name': "DR - RIO DE JANEIRO"},
    60: {'code': "RN", 'name': "DR - RIO GRANDE DO NORTE"},
    26: {'code': "RO", 'name': "DR - RONDONIA"},
    65: {'code': "RR", 'name': "DR - RORAIMA"},
    64: {'code': "RS", 'name': "DR - RIO GRANDE DO SUL"},
    68: {'code': "SC", 'name': "DR - SANTA CATARINA"},
    70: {'code': "SE", 'name': "DR - SERGIPE"},
    74: {'code': "SPI", 'name': "DR - SÃO PAULO INTERIOR"},
    72: {'code': "SPM", 'name': "DR - SÃO PAULO"},
    75: {'code': "TO", 'name': "DR - TOCANTINS"},
}

TRACKING_EVENT_TYPES = {
    "ERROR": "Evento de erro",  # custom event type for "Not Found" error
    "BDE": "Baixa de distribuição externa",
    "BDI": "Baixa de distribuição interna",
    "BDR": "Baixa corretiva",
    "BLQ": "Bloqueio de objetos",
    "CAR": "Conferência de lista de registro",
    "CD": "Conferência de nota de despacho",
    "CMT": "Chegada de um meio de transporte",
    "CO": "Coleta de objetos",
    "CUN": "Conferência de lista de registro",
    "DO": "Expedição de nota de despacho",
    "EST": "Estorno",
    "FC": "Função complementar",
    "IDC": "Indenização de objetos",
    "LDI": "Lista de distribuição interna",
    "LDE": "Lista de distribuição externa",
    "OEC": "Lista de Objetos Entregues ao Carteiro",
    "PAR": "Conferência Unidade Internacional",
    "PMT": "Partida Meio de Transporte",
    "PO": "Postagem (exceção)",
    "RO": "Expedição de Lista de Registro",
    "TRI": "Triagem",
}



# Copyright 2016 Osvaldo Santana Neto
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import math
import os
from datetime import datetime
from decimal import Decimal
from typing import Optional, Sequence, Tuple, Union

from PIL import Image

from correios import DATADIR
from correios.exceptions import (InvalidAddressesError, InvalidEventStatusError,
                                 InvalidPackageSequenceError, InvalidTrackingCodeError,
                                 PostingListError, InvalidPackageDimensionsError,
                                 InvalidPackageWeightError)
from .address import Address, ZipCode
from .data import SERVICE_PAC, TRACKING_EVENT_TYPES
from .user import Contract  # noqa: F401
from .user import Service, ExtraService, PostingCard

TRACKING_CODE_SIZE = 13
TRACKING_CODE_NUMBER_SIZE = 8
TRACKING_CODE_PREFIX_SIZE = 2
TRACKING_CODE_SUFFIX_SIZE = 2
IATA_COEFICIENT = 6.0
VOLUMETRIC_WEIGHT_THRESHOLD = 10000  # g
MIN_WIDTH, MAX_WIDTH = 11, 105  # cm
MIN_HEIGHT, MAX_HEIGHT = 2, 105  # cm
MIN_LENGTH, MAX_LENGTH = 16, 105  # cm
MIN_DIAMETER, MAX_DIAMETER = 16, 91  # cm
MIN_CYLINDER_LENGTH, MAX_CYLINDER_LENGTH = 18, 105  # cm
MIN_SIZE, MAX_SIZE = 29, 200  # cm
MAX_CYLINDER_SIZE = 28
INSURANCE_VALUE_THRESHOLD = 50  # R$


class EventStatus:
    def __init__(self, event_type: str, status: int):
        self.type = self._validate_type(event_type)
        self.status = status

    def _validate_type(self, event_type):
        event_type = event_type.upper()

        if event_type not in TRACKING_EVENT_TYPES:
            raise InvalidEventStatusError("{} is not valid".format(event_type))

        return event_type

    @property
    def display_event_type(self):
        return TRACKING_EVENT_TYPES[self.type]

    def __str__(self):
        return '({}, {})'.format(self.type, self.status)

    def __repr__(self):
        return '<EventStatus({!r}, {!r})>'.format(self.type, self.status)


class TrackingEvent:
    timestamp_format = "%d/%m/%Y %H:%M"

    def __init__(self,
                 timestamp: datetime,
                 status: Union[Tuple[str, int], EventStatus],
                 location_zip_code: Union[str, ZipCode] = "",
                 location: str = "",
                 receiver: str = "",
                 city: str = "",
                 state: str = "",
                 document: str = "",
                 comment: str = "",
                 description: str = "",
                 details: str = "",
                 ):
        self.timestamp = timestamp
        self.location = location
        self.receiver = receiver
        self.city = city
        self.state = state
        self.document = document
        self.comment = comment
        self.description = description
        self.details = details

        if location_zip_code:
            location_zip_code = ZipCode.create(location_zip_code)
        self.location_zip_code = location_zip_code

        if isinstance(status, tuple):
            status = EventStatus(*status)
        self.status = status

    def __str__(self):
        return '{} - {} - {}/{}'.format(self.description, self.location, self.city, self.state)

    def __repr__(self):
        timestamp = self.timestamp.strftime(self.timestamp_format)
        return '<TrackingEvent({!s}, {!s})>'.format(self.status, timestamp)


class NotFoundTrackingEvent(TrackingEvent):
    def __init__(self,
                 timestamp: datetime,
                 status: Union[Tuple[str, int], EventStatus],
                 comment,
                 ):
        super().__init__(timestamp=timestamp,
                         status=status,
                         comment=comment)


class TrackingCode:
    def __init__(self, code: str):
        self.prefix = code[:2].upper()
        self.number = "".join(d for d in code[2:10] if d.isdigit())
        self.suffix = code[-2:].upper()
        self._digit = None

        if len(code) == TRACKING_CODE_SIZE and code[10:11] != " ":
            self._digit = int(code[10:11])

        self._validate()

        # filled by tracking service
        self.category = None
        self.name = None
        self.initials = None
        self.events = []

    def _validate(self):
        if len(self.prefix) != TRACKING_CODE_PREFIX_SIZE or not self.prefix.isalpha():
            raise InvalidTrackingCodeError("Invalid tracking code prefix {}".format(self.prefix))

        if len(self.suffix) != TRACKING_CODE_SUFFIX_SIZE or not self.suffix.isalpha():
            raise InvalidTrackingCodeError("Invalid tracking code suffix {}".format(self.suffix))

        if len(self.number) != TRACKING_CODE_NUMBER_SIZE or not self.number.isnumeric():
            raise InvalidTrackingCodeError("Invalid tracking code number {}".format(self.number))

        if self._digit is not None and self._digit != self.calculate_digit(self.number):
            raise InvalidTrackingCodeError("Invalid tracking code number {} or digit {} (must be {})".format(
                self.number,
                self._digit,
                self.calculate_digit(self.number))
            )

    @classmethod
    def create(cls, tracking_code: Union[str, 'TrackingCode']):
        if isinstance(tracking_code, cls):
            return tracking_code
        return cls(tracking_code)

    @classmethod
    def calculate_digit(cls, number: str) -> int:
        numbers = [int(c) for c in number if c.isdigit()]

        multipliers = [8, 6, 4, 2, 3, 5, 9, 7]
        module = sum(multipliers[i] * digit for i, digit in enumerate(numbers)) % 11

        if not module:
            return 5

        if module == 1:
            return 0

        return 11 - module

    @classmethod
    def create_range(cls, start: Union[str, 'TrackingCode'], end: Union[str, 'TrackingCode']):
        if not isinstance(start, TrackingCode):
            start = TrackingCode(start)

        if not isinstance(end, TrackingCode):
            end = TrackingCode(end)

        if start.prefix != end.prefix:
            raise InvalidTrackingCodeError("Different tracking code prefixes: {} != {}".format(start.prefix,
                                                                                               end.prefix))

        if start.suffix != end.suffix:
            raise InvalidTrackingCodeError("Different tracking code suffixes: {} != {}".format(start.suffix,
                                                                                               end.suffix))

        start_number = int(start.number)
        end_number = int(end.number)

        if start_number > end_number:
            raise InvalidTrackingCodeError("Invalid range numbers: {} > {}".format(start_number,
                                                                                   end_number))

        code_range = range(int(start.number), int(end.number) + 1)
        return [TrackingCode(start.prefix + "{:08}".format(n) + start.suffix) for n in code_range]

    @property
    def digit(self):
        if self._digit is None:
            self._digit = self.calculate_digit(self.number)
        return self._digit

    @property
    def code(self):
        return self.prefix + self.number + str(self.digit) + self.suffix

    @property
    def nodigit(self):
        return "{}{} {}".format(self.prefix, self.number, self.suffix)

    @property
    def short(self):
        return "{}{}{}".format(self.prefix, self.number, self.suffix)

    @property
    def splitted(self):
        code = self.code
        return "{!s} {!s} {!s} {!s} {!s}".format(code[:2], code[2:5], code[5:8], code[8:11], code[11:])

    def add_event(self, event: TrackingEvent):
        self.events.append(event)

    def __str__(self):
        return self.code

    def __repr__(self):
        return "<TrackingCode code={!r}>".format(self.code)


class Package:
    TYPE_ENVELOPE = 1
    TYPE_BOX = 2
    TYPE_CYLINDER = 3

    def __init__(self,
                 package_type: int = TYPE_BOX,
                 width: int = 0, height: int = 0, length: int = 0, diameter: int = 0, weight: int = 0,
                 sequence=(1, 1),
                 service: Optional['Service'] = None):

        Package.validate(package_type, width, height, length, diameter, service, weight)

        if len(sequence) != 2 or sequence[0] > sequence[1]:
            raise InvalidPackageSequenceError("Package must be a tuple with 2 elements: (number, total)")

        self.package_type = package_type
        self.width = width  # cm
        self.height = height  # cm
        self.length = length  # cm
        self.diameter = diameter  # in cm
        self.weight = weight  # in grams
        self.sequence = sequence
        self.service = service

    @property
    def volumetric_weight(self):
        return Package.calculate_volumetric_weight(self.width, self.height, self.length)

    @property
    def posting_weight(self):
        return Package.calculate_posting_weight(self.weight, self.volumetric_weight)

    @classmethod
    def calculate_volumetric_weight(cls, width, height, length):
        return math.ceil((width * height * length) / IATA_COEFICIENT)

    @classmethod
    def calculate_posting_weight(cls, weight, volumetric_weight):
        if volumetric_weight <= VOLUMETRIC_WEIGHT_THRESHOLD:
            return weight
        return math.ceil(max(volumetric_weight, weight))

    @classmethod
    def calculate_insurance(cls, per_unit_value, quantity: int = 1, service: Service = None):
        value = 0
        if service == SERVICE_PAC and per_unit_value > INSURANCE_VALUE_THRESHOLD:
            value = float(per_unit_value - INSURANCE_VALUE_THRESHOLD) * 0.007

        return Decimal(value * quantity).quantize(Decimal('0.00'))

    @classmethod
    def validate(cls, package_type: int, width: int = 0, height: int = 0, length: int = 0, diameter: int = 0,
                 service: Optional[Service] = None, weight: int = 0):

        if service and service.max_weight and weight > service.max_weight:
            raise InvalidPackageWeightError("Max weight exceeded {!r}g (max. {!r}g)".format(weight, service.max_weight))

        if package_type == Package.TYPE_ENVELOPE:
            if any([width, height, length, diameter]):
                raise InvalidPackageDimensionsError("Invalid dimensions: {}x{}x{}".format(width, height, length))
            return

        if package_type == Package.TYPE_BOX:
            if diameter:
                raise InvalidPackageDimensionsError("Package does not use diameter: {}".format(diameter))

            if not MIN_WIDTH <= width <= MAX_WIDTH:
                raise InvalidPackageDimensionsError("Invalid width (range 11~105): {}".format(width))

            if not MIN_HEIGHT <= height <= MAX_HEIGHT:
                raise InvalidPackageDimensionsError("Invalid height (range 2~105): {}".format(height))

            if not MIN_LENGTH <= length <= MAX_LENGTH:
                raise InvalidPackageDimensionsError("Invalid length (range 16~105): {}".format(length))

            if not MIN_SIZE <= (width + height + length) <= MAX_SIZE:
                raise InvalidPackageDimensionsError("Invalid dimensions: {}x{}x{}".format(width, height, length))

            return

        # Volume.TYPE_CYLINDER
        if width or height:
            raise InvalidPackageDimensionsError("Cylinder does not use width/height: {}x{}".format(width, height))

        if not MIN_CYLINDER_LENGTH <= length <= MAX_CYLINDER_LENGTH:
            raise InvalidPackageDimensionsError("Invalid length (range 18~105): {}".format(length))

        if not MIN_DIAMETER <= diameter <= MAX_DIAMETER:
            raise InvalidPackageDimensionsError("Invalid diameter (range 5~91): {}".format(diameter))

        if (length + 2 * diameter) > MAX_CYLINDER_SIZE:
            raise InvalidPackageDimensionsError("Invalid dimensions: {}x{}".format(length, diameter))


class ShippingLabel:
    variable_data_identifier = 51  # Variable data identifier for package
    invoice_template = "{!s}"
    contract_number_template = "{!s}"
    order_template = "{!s}"
    service_name_template = "{!s}"
    package_template = "{!s}/{!s}"
    weight_template = "{!s}g"
    receipt_template = ("Recebedor: ___________________________________________<br/>"
                        "Assinatura: __________________ Documento: _______________")
    sender_header = "DESTINATÁRIO"
    carrier_logo = os.path.join(DATADIR, "carrier_logo_bw.png")
    receiver_data_template = ("{receiver.name}<br/>"
                              "{receiver.street}, {receiver.number}<br/>"
                              "{receiver.complement} {receiver.neighborhood}<br/>"
                              "<b>{receiver.zip_code_display}</b> {receiver.city}/{receiver.state}")

    sender_data_template = ("<b>Remetente:</b> {sender.name}<br/>"
                            "{sender.street}, {sender.number}<br/>"
                            "{sender.complement} - {sender.neighborhood}<br/>"
                            "<b>{sender.zip_code_display}</b> {sender.city}-{sender.state}")

    def __init__(self,
                 posting_card: PostingCard,
                 sender: Address,
                 receiver: Address,
                 service: Union[Service, int],
                 tracking_code: Union[TrackingCode, str],
                 package: Package,
                 extra_services: Optional[Sequence[Union[ExtraService, int]]] = None,
                 logo: Optional[Union[str, Image.Image]] = None,
                 order: Optional[str] = "",
                 invoice_number: Optional[str] = "",
                 invoice_series: Optional[str] = "",
                 invoice_type: Optional[str] = "",
                 value: Optional[Decimal] = Decimal("0.00"),
                 billing: Optional[Decimal] = Decimal("0.00"),
                 text: Optional[str] = "",
                 latitude: Optional[float] = 0.0,
                 longitude: Optional[float] = 0.0):

        if sender == receiver:
            raise InvalidAddressesError("Sender and receiver cannot be the same")

        if logo is None:
            logo = os.path.join(DATADIR, "default_logo.png")

        if isinstance(logo, str):
            logo = Image.open(logo)

        self.posting_card = posting_card
        self.sender = sender
        self.receiver = receiver
        self.service = Service.get(service)
        self.tracking_code = TrackingCode.create(tracking_code)
        self.package = package
        self.logo = logo
        self.order = order
        self.invoice_number = invoice_number
        self.invoice_series = invoice_series
        self.invoice_type = invoice_type
        self.value = value
        self.billing = billing
        self.text = text
        self.latitude = latitude
        self.longitude = longitude
        self.carrier_logo = Image.open(self.carrier_logo)

        self.extra_services = self.service.default_extra_services[:]
        if extra_services:
            self.extra_services.extend(ExtraService.get(es) for es in extra_services)

        self.posting_list = None
        self.posting_list_group = 0

    def __repr__(self):
        return "<ShippingLabel tracking={!r}>".format(str(self.tracking_code))

    @property
    def symbol(self):
        return self.service.symbol_image

    @property
    def contract(self):
        return self.posting_card.contract

    @property
    def posting_weight(self):
        return self.package.posting_weight

    def get_order(self):
        return self.order_template.format(self.order)

    def get_invoice(self):
        return self.invoice_template.format(self.invoice_number)

    def get_contract_number(self):
        return self.contract_number_template.format(self.posting_card.get_contract_number())

    def get_service_name(self):
        return self.service_name_template.format(self.service.display_name)

    def get_package_sequence(self):
        return self.package_template.format(*self.package.sequence)

    def get_weight(self):
        return self.weight_template.format(self.package.weight)

    def get_symbol_filename(self, extension="gif"):
        return self.service.get_symbol_filename(extension)

    def get_tracking_code(self):
        return self.tracking_code.splitted

    def get_receiver_data(self):
        return self.receiver_data_template.format(receiver=self.receiver)

    def get_sender_data(self):
        return self.sender_data_template.format(sender=self.sender)

    def _get_extra_service_info(self) -> str:
        extra_services_numbers = ["00" for _ in range(6)]
        for i, extra_service in enumerate(self.extra_services[:6]):
            extra_services_numbers[i] = "{:02d}".format(extra_service.number)
        return "".join(extra_services_numbers)

    def get_datamatrix_info(self):
        parts = [
            "{!s:>08}".format(self.receiver.zip_code),
            "{!s:>05}".format(self.receiver.number),
            "{!s:>08}".format(self.sender.zip_code),
            "{!s:>05}".format(self.sender.number),
            "{!s:>01}".format(self.receiver.zip_code.digit),
            "{!s:>02}".format(self.variable_data_identifier),
            "{!s:>13}".format(self.tracking_code),
            "{!s:>12}".format(self._get_extra_service_info()),
            "{!s:>010}".format(self.posting_card.number),
            "{!s:>05}".format(self.service.code),
            "{!s:>02}".format(self.posting_list_group),
            "{!s:>05}".format(self.receiver.number),
            "{!s:<20}".format(self.receiver.complement[:20]),
            "{!s:>05}".format(str(int(self.value * 100))),
            "{!s:>012}".format(str(self.receiver.phone)[:12] or "0" * 12),
            "{:+010.6f}".format(self.latitude),
            "{:+010.6f}".format(self.longitude),
            "|",
            "{!s:<30}".format(self.text[:30]),
        ]
        return "".join(parts)

    def __contains__(self, extra_service: ExtraService):
        return extra_service in self.extra_services


class PostingList:
    def __init__(self, custom_id: int, logo: Optional[Union[str, Image.Image]] = None):
        self.number = None  # will be filled by close_posting_list

        if logo is None:
            logo = os.path.join(DATADIR, "carrier_logo.png")

        if isinstance(logo, str):
            logo = Image.open(logo)

        self.logo = logo
        self.custom_id = custom_id
        self.shipping_labels = {}

        # filled by the first shipping label
        self.initial_shipping_label = None
        self.posting_card = None  # type: PostingCard
        self.contract = None  # type: Contract
        self.sender = None  # type: Address

    def add_shipping_label(self, shipping_label: ShippingLabel):
        if not self.initial_shipping_label:
            self.initial_shipping_label = shipping_label
            self.posting_card = shipping_label.posting_card
            self.contract = shipping_label.contract
            self.sender = shipping_label.sender

        if shipping_label.tracking_code.short in self.shipping_labels:
            raise PostingListError("Shipping label {!r} already in posting list".format(shipping_label))

        if shipping_label.posting_card != self.posting_card:
            raise PostingListError("Invalid posting card: {} != {}".format(shipping_label.posting_card,
                                                                           self.posting_card))
        self.shipping_labels[shipping_label.tracking_code.short] = shipping_label
        shipping_label.posting_list = self

    def get_tracking_codes(self):
        return list(self.shipping_labels.keys())

    def close_with_id(self, number: int):
        self.number = number

    @property
    def closed(self):
        return self.number is not None
